@@ -21,87 +21,91 @@
 
 using namespace UCOMMON_NAMESPACE;
 
-SharedProtocol::~SharedProtocol()
+UnlockAccess::~UnlockAccess()
 {
 }
 
-ExclusiveProtocol::~ExclusiveProtocol()
+SharedAccess::~SharedAccess()
 {
 }
 
-void SharedProtocol::Exclusive(void)
+ExclusiveAccess::~ExclusiveAccess()
 {
 }
 
-void SharedProtocol::Share(void)
+void SharedAccess::exclusive(void)
 {
 }
 
-shared_lock::shared_lock(SharedProtocol *obj)
+void SharedAccess::share(void)
+{
+}
+
+shared_access::shared_access(SharedAccess *obj)
 {
     assert(obj != NULL);
     lock = obj;
     modify = false;
-    lock->Shlock();
+    lock->shared_lock();
 }
 
-exclusive_lock::exclusive_lock(ExclusiveProtocol *obj)
+exclusive_access::exclusive_access(ExclusiveAccess *obj)
 {
     assert(obj != NULL);
     lock = obj;
-    lock->Exlock();
+    lock->exclusive_lock();
 }
 
-shared_lock::~shared_lock()
+shared_access::~shared_access()
 {
     if(lock) {
         if(modify)
-            lock->Share();
-        lock->Unlock();
+            lock->share();
+        lock->release_share();
         lock = NULL;
         modify = false;
     }
 }
 
-exclusive_lock::~exclusive_lock()
+exclusive_access::~exclusive_access()
 {
     if(lock) {
-        lock->Unlock();
+        lock->release_exclusive();
         lock = NULL;
     }
 }
 
-void shared_lock::release()
+void shared_access::release()
 {
     if(lock) {
         if(modify)
-            lock->Share();
-        lock->Unlock();
+            lock->share();
+        lock->release_share();
         lock = NULL;
         modify = false;
     }
 }
 
-void exclusive_lock::release()
+void exclusive_access::release()
 {
     if(lock) {
-        lock->Unlock();
+        lock->release_exclusive();
         lock = NULL;
     }
 }
 
-void shared_lock::exclusive(void)
+void shared_access::exclusive(void)
 {
     if(lock && !modify) {
-        lock->Exclusive();
+        lock->exclusive();
         modify = true;
     }
 }
 
-void shared_lock::share(void)
+void shared_access::share(void)
 {
     if(lock && modify) {
-        lock->Share();
+        lock->share();
         modify = false;
     }
 }
