@@ -57,6 +57,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
+#include <ctype.h>
 
 #ifdef  HAVE_POSIX_FADVISE
 #ifndef POSIX_FADV_RANDOM
@@ -85,9 +86,6 @@
 using namespace UCOMMON_NAMESPACE;
 
 const fsys::offset_t fsys::end = (offset_t)(-1);
-charfile cstdin(stdin);
-charfile cstdout(stdout);
-charfile cstderr(stderr);
 
 #ifdef  _MSWINDOWS_
 
@@ -169,27 +167,31 @@ int fsys::remapError(void)
     case ERROR_SIGNAL_PENDING:
     case ERROR_BUSY:
         return EBUSY;
+    case ERROR_DIR_NOT_EMPTY:
+        return ENOTEMPTY;
+    case ERROR_DIRECTORY:
+        return ENOTDIR;
     default:
         return EINVAL;
     }
 }
 
-int fsys::createDir(const char *path, unsigned mode)
+int dir::create(const char *path, unsigned perms)
 {
     if(!CreateDirectory(path, NULL))
         return remapError();
 
-    return change(path, mode);
+    if(perms & 06)
+        perms |= 01;
+    if(perms & 060)
+        perms |= 010;
+    if(perms & 0600)
+        perms |= 0100;
+
+    return mode(path, perms);
 }
 
-int fsys::removeDir(const char *path)
-{
-    if(_rmdir(path))
-        return remapError();
-    return 0;
-}
-
-fd_t fsys::nullfile(void)
+fd_t fsys::null(void)
 {
     SECURITY_ATTRIBUTES sattr;
 
@@ -215,7 +217,7 @@ int fsys::pipe(fd_t& input, fd_t& output, size_t size)
     return 0;
 }
 
-int fsys::fileinfo(const char *path, struct stat *buf)
+int fsys::info(const char *path, struct stat *buf)
 {
     if(_stat(path, (struct _stat *)(buf)))
         return remapError();
@@ -233,7 +235,7 @@ int fsys::trunc(offset_t offset)
     return remapError();
 }
 
-int fsys::fileinfo(struct stat *buf)
+int fsys::info(struct stat *buf)
 {
     int fn = _open_osfhandle((intptr_t)(fd), O_RDONLY);
 
@@ -244,35 +246,74 @@ int fsys::fileinfo(struct stat *buf)
     return rtn;
 }
 
-int fsys::changeDir(const char *path)
+int fsys::prefix(const char *path)
 {
     if (_chdir(path))
         return remapError();
     return 0;
 }
 
-int fsys::getPrefix(char *path, size_t len)
+int fsys::prefix(char *path, size_t len)
 {
     if (_getcwd(path, len))
         return remapError();
     return 0;
 }
 
-int fsys::change(const char *path, unsigned mode)
+int fsys::mode(const char *path, unsigned value)
 {
-    if(_chmod(path, mode))
+    if(_chmod(path, value))
         return remapError();
     return 0;
 }
 
-int fsys::access(const char *path, unsigned mode)
+bool fsys::is_exists(const char *path)
 {
-    if(_access(path, mode))
-        return remapError();
-    return 0;
+    if(_access(path, F_OK))
+        return false;
+    return true;
 }
 
-bool fsys::istty(fd_t fd)
+bool fsys::is_readable(const char *path)
+{
+    if(_access(path, R_OK))
+        return false;
+    return true;
+}
+
+bool fsys::is_writable(const char *path)
+{
+    if(_access(path, W_OK))
+        return false;
+    return true;
+}
+
+bool fsys::is_executable(const char *path)
+{
+    path = strrchr(path, '.');
+
+    if(!path)
+        return false;
+
+    if(eq_case(path, ".exe"))
+        return true;
+
+    if(eq_case(path, ".bat"))
+        return true;
+
+    if(eq_case(path, ".com"))
+        return true;
+
+    if(eq_case(path, ".cmd"))
+        return true;
+
+    if(eq_case(path, ".ps1"))
+        return true;
+
+    return false;
+}
+
+bool fsys::is_tty(fd_t fd)
 {
     if(fd == INVALID_HANDLE_VALUE)
         return false;
@@ -282,7 +323,7 @@ bool fsys::istty(fd_t fd)
     return false;
 }
 
-bool fsys::istty(void)
+bool fsys::is_tty(void)
 {
     error = 0;
     if(fd == INVALID_HANDLE_VALUE)
@@ -295,11 +336,9 @@ bool fsys::istty(void)
     return false;
 }
 
-void fsys::close(void)
+void dir::close(void)
 {
     error = 0;
-    if(fd == INVALID_HANDLE_VALUE)
-        return;
 
     if(ptr) {
         if(::FindClose(fd)) {
@@ -310,17 +349,27 @@ void fsys::close(void)
         else
             error = remapError();
     }
-    else if(::CloseHandle(fd))
+    else
+        error = EBADF;
+}
+
+int fsys::close(void)
+{
+    error = 0;
+
+    if(fd == INVALID_HANDLE_VALUE)
+        return EBADF;
+
+    if(::CloseHandle(fd))
         fd = INVALID_HANDLE_VALUE;
     else
         error = remapError();
+    return error;
 }
 
-ssize_t fsys::read(void *buf, size_t len)
+ssize_t dir::read(char *buf, size_t len)
 {
     ssize_t rtn = -1;
-    DWORD count;
-
     if(ptr) {
         snprintf((char *)buf, len, ptr->cFileName);
         rtn = strlen(ptr->cFileName);
@@ -328,6 +377,13 @@ ssize_t fsys::read(void *buf, size_t len)
             close();
         return rtn;
     }
+    return -1;
+}
+
+ssize_t fsys::read(void *buf, size_t len)
+{
+    ssize_t rtn = -1;
+    DWORD count;
 
     if(ReadFile(fd, (LPVOID) buf, (DWORD)len, &count, NULL))
         rtn = count;
@@ -398,60 +454,102 @@ void fsys::release(fd_t fd)
     CloseHandle(fd);
 }
 
+void dir::open(const char *path)
+{
+    close();
+
+    char tpath[256];
+    DWORD attr = GetFileAttributes(path);
+
+    if((attr == (DWORD)~0l) || !(attr & FILE_ATTRIBUTE_DIRECTORY)) {
+        error = ENOTDIR;
+        return;
+    }
+
+    snprintf(tpath, sizeof(tpath), "%s%s", path, "\\*");
+    ptr = new WIN32_FIND_DATA;
+    fd = FindFirstFile(tpath, ptr);
+    if(fd == INVALID_HANDLE_VALUE) {
+        delete ptr;
+        ptr = NULL;
+        error = remapError();
+    }
+    return;
+}
+
 void fsys::open(const char *path, access_t access)
 {
     bool append = false;
-    DWORD amode;
+    DWORD amode = 0;
     DWORD smode = 0;
     DWORD attr = FILE_ATTRIBUTE_NORMAL;
+    char buf[128];
 
     close();
 
+    if(access == DEVICE) {
+#ifdef  _MSWINDOWS_
+        if(isalpha(path[0]) && path[1] == ':') {
+            if(!QueryDosDevice(path, buf, sizeof(buf))) {
+                error = ENODEV;
+                return;
+            }
+            path = buf;
+        }
+#else
+        if(!strchr(path, '/')) {
+            if(path[0] == 'S' && isdigit(path[1]))
+                snprintf(buf, sizeof(buf) "/dev/tty%s", path);
+            else if(strncmp(path, "USB", 3))
+                snprintf(buf, sizeof(buf), "/dev/tty%s", path);
+            else
+                snprintf(buf, sizeof(buf), "/dev/%s", path);
+
+            char *cp = strchr(buf, ':');
+            if(cp)
+                *cp = 0;
+            path = buf;
+        }
+#endif
+        if(!is_device(buf)) {
+            error = ENODEV;
+            return;
+        }
+    }
+
     switch(access)
     {
-    case ACCESS_STREAM:
+    case STREAM:
 #ifdef  FILE_FLAG_SEQUENTIAL_SCAN
         attr |= FILE_FLAG_SEQUENTIAL_SCAN;
 #endif
-    case ACCESS_RDONLY:
+    case RDONLY:
         amode = GENERIC_READ;
         smode = FILE_SHARE_READ;
         break;
-    case ACCESS_WRONLY:
+    case WRONLY:
         amode = GENERIC_WRITE;
         break;
-    case ACCESS_RANDOM:
+    case DEVICE:
+        smode = FILE_SHARE_READ;
+        attr |= FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING;
+    case EXCLUSIVE:
+        amode = GENERIC_READ | GENERIC_WRITE;
+        break;
+    case RANDOM:
         attr |= FILE_FLAG_RANDOM_ACCESS;
-    case ACCESS_REWRITE:
+    case REWRITE:
         amode = GENERIC_READ | GENERIC_WRITE;
         smode = FILE_SHARE_READ;
         break;
-    case ACCESS_APPEND:
+    case APPEND:
         amode = GENERIC_WRITE;
         append = true;
         break;
-    case ACCESS_SHARED:
+    case SHARED:
         amode = GENERIC_READ | GENERIC_WRITE;
         smode = FILE_SHARE_READ | FILE_SHARE_WRITE;
         break;
-    case ACCESS_DIRECTORY:
-        char tpath[256];
-        DWORD attr = GetFileAttributes(path);
-
-        if((attr == (DWORD)~0l) || !(attr & FILE_ATTRIBUTE_DIRECTORY)) {
-            error = ENOTDIR;
-            return;
-        }
-
-        snprintf(tpath, sizeof(tpath), "%s%s", path, "\\*");
-        ptr = new WIN32_FIND_DATA;
-        fd = FindFirstFile(tpath, ptr);
-        if(fd == INVALID_HANDLE_VALUE) {
-            delete ptr;
-            ptr = NULL;
-            error = remapError();
-        }
-        return;
     }
 
     fd = CreateFile(path, amode, smode, NULL, OPEN_EXISTING, attr, NULL);
@@ -461,14 +559,16 @@ void fsys::open(const char *path, access_t access)
         error = remapError();
 }
 
-void fsys::create(const char *path, access_t access, unsigned mode)
+void fsys::open(const char *path, unsigned fmode, access_t access)
 {
     bool append = false;
-    DWORD amode;
-    DWORD cmode;
+    DWORD amode = 0;
+    DWORD cmode = 0;
     DWORD smode = 0;
     DWORD attr = FILE_ATTRIBUTE_NORMAL;
 
+    fmode &= 0666;
+
     const char *cp = strrchr(path, '\\');
     const char *cp2 = strrchr(path, '/');
     if(cp2 > cp)
@@ -484,37 +584,41 @@ void fsys::create(const char *path, access_t access, unsigned mode)
 
     switch(access)
     {
-    case ACCESS_RDONLY:
+    case DEVICE:
+        error = ENOSYS;
+        return;
+
+    case RDONLY:
         amode = GENERIC_READ;
         cmode = OPEN_ALWAYS;
         smode = FILE_SHARE_READ;
         break;
-    case ACCESS_STREAM:
-    case ACCESS_WRONLY:
+    case STREAM:
+    case WRONLY:
         amode = GENERIC_WRITE;
         cmode = CREATE_ALWAYS;
         break;
-    case ACCESS_RANDOM:
+    case EXCLUSIVE:
+        amode = GENERIC_READ | GENERIC_WRITE;
+        cmode = OPEN_ALWAYS;
+        break;
+    case RANDOM:
         attr |= FILE_FLAG_RANDOM_ACCESS;
-    case ACCESS_REWRITE:
+    case REWRITE:
         amode = GENERIC_READ | GENERIC_WRITE;
         cmode = OPEN_ALWAYS;
         smode = FILE_SHARE_READ;
         break;
-    case ACCESS_APPEND:
+    case APPEND:
         amode = GENERIC_WRITE;
         cmode = OPEN_ALWAYS;
         append = true;
         break;
-    case ACCESS_SHARED:
+    case SHARED:
         amode = GENERIC_READ | GENERIC_WRITE;
         cmode = OPEN_ALWAYS;
         smode = FILE_SHARE_READ | FILE_SHARE_WRITE;
         break;
-    case ACCESS_DIRECTORY:
-        createDir(path, mode);
-        open(path, access);
-        return;
     }
     fd = CreateFile(path, amode, smode, NULL, cmode, attr, NULL);
     if(fd != INVALID_HANDLE_VALUE && append)
@@ -522,14 +626,13 @@ void fsys::create(const char *path, access_t access, unsigned mode)
     else if(fd == INVALID_HANDLE_VALUE)
         error = remapError();
     if(fd != INVALID_HANDLE_VALUE)
-        change(path, mode);
+        mode(path, fmode);
 }
 
 fsys::fsys(const fsys& copy)
 {
     error = 0;
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
 
     if(copy.fd == INVALID_HANDLE_VALUE)
         return;
@@ -621,7 +724,7 @@ int fsys::seek(offset_t pos)
 
 #else
 
-ssize_t fsys::read(void *buf, size_t len)
+ssize_t dir::read(char *buf, size_t len)
 {
     if(ptr) {
         dirent *entry = ::readdir((DIR *)ptr);
@@ -632,7 +735,11 @@ ssize_t fsys::read(void *buf, size_t len)
         String::set((char *)buf, len, entry->d_name);
         return strlen(entry->d_name);
     }
+    return -1;
+}
 
+ssize_t fsys::read(void *buf, size_t len)
+{
 #ifdef  __PTH__
     int rtn = ::pth_read(fd, buf, len);
 #else
@@ -646,11 +753,6 @@ ssize_t fsys::read(void *buf, size_t len)
 
 int fsys::sync(void)
 {
-    if(ptr) {
-        error = EBADF;
-        return -1;
-    }
-
     int rtn = ::fsync(fd);
     if(rtn < 0)
         error = remapError();
@@ -661,11 +763,6 @@ int fsys::sync(void)
 
 ssize_t fsys::write(const void *buf, size_t len)
 {
-    if(ptr) {
-        error = EBADF;
-        return -1;
-    }
-
 #ifdef  __PTH__
     int rtn = pth_write(fd, buf, len);
 #else
@@ -677,7 +774,7 @@ ssize_t fsys::write(const void *buf, size_t len)
     return rtn;
 }
 
-fd_t fsys::nullfile(void)
+fd_t fsys::null(void)
 {
     return ::open("/dev/null", O_RDWR);
 }
@@ -693,21 +790,21 @@ int fsys::pipe(fd_t& input, fd_t& output, size_t size)
     return 0;
 }
 
-bool fsys::istty(fd_t fd)
+bool fsys::is_tty(fd_t fd)
 {
     if(isatty(fd))
         return true;
     return false;
 }
 
-bool fsys::istty(void)
+bool fsys::is_tty(void)
 {
     if(isatty(fd))
         return true;
     return false;
 }
 
-void fsys::close(void)
+void dir::close(void)
 {
     error = 0;
     if(ptr) {
@@ -715,12 +812,22 @@ void fsys::close(void)
             error = remapError();
         ptr = NULL;
     }
-    else if(fd != INVALID_HANDLE_VALUE) {
+    else
+        error = EBADF;
+}
+
+int fsys::close(void)
+{
+    error = 0;
+    if(fd != INVALID_HANDLE_VALUE) {
         if(::close(fd) == 0)
             fd = INVALID_HANDLE_VALUE;
         else
             error = remapError();
     }
+    else
+        error = EBADF;
+    return error;
 }
 
 fd_t fsys::input(const char *path)
@@ -730,12 +837,12 @@ fd_t fsys::input(const char *path)
 
 fd_t fsys::output(const char *path)
 {
-    return ::open(path, O_WRONLY | O_CREAT | O_TRUNC, 0666);
+    return ::open(path, O_WRONLY | O_CREAT | O_TRUNC, EVERYONE);
 }
 
 fd_t fsys::append(const char *path)
 {
-    return ::open(path, O_WRONLY | O_CREAT | O_APPEND, 0666);
+    return ::open(path, O_WRONLY | O_CREAT | O_APPEND, EVERYONE);
 }
 
 void fsys::release(fd_t fd)
@@ -743,7 +850,7 @@ void fsys::release(fd_t fd)
     ::close(fd);
 }
 
-void fsys::create(const char *path, access_t access, unsigned mode)
+void fsys::open(const char *path, unsigned fmode, access_t access)
 {
     unsigned flags = 0;
 
@@ -751,49 +858,59 @@ void fsys::create(const char *path, access_t access, unsigned mode)
 
     switch(access)
     {
-    case ACCESS_RDONLY:
+    case DEVICE:
+        error = ENOSYS;
+        return;
+
+    case RDONLY:
         flags = O_RDONLY | O_CREAT;
         break;
-    case ACCESS_STREAM:
-    case ACCESS_WRONLY:
+    case STREAM:
+    case WRONLY:
         flags = O_WRONLY | O_CREAT | O_TRUNC;
         break;
-    case ACCESS_RANDOM:
-    case ACCESS_SHARED:
-    case ACCESS_REWRITE:
+    case RANDOM:
+    case SHARED:
+    case REWRITE:
+    case EXCLUSIVE:
         flags = O_RDWR | O_CREAT;
         break;
-    case ACCESS_APPEND:
+    case APPEND:
         flags = O_RDWR | O_APPEND | O_CREAT;
         break;
-    case ACCESS_DIRECTORY:
-        ::mkdir(path, mode);
-        open(path, access);
-        return;
     }
-    fd = ::open(path, flags, mode);
+    fd = ::open(path, flags, fmode);
     if(fd == INVALID_HANDLE_VALUE)
         error = remapError();
 #ifdef HAVE_POSIX_FADVISE
     else {
-        if(access == ACCESS_RANDOM)
+        if(access == RANDOM)
             posix_fadvise(fd, (off_t)0, (off_t)0, POSIX_FADV_RANDOM);
     }
 #endif
 }
 
-int fsys::createDir(const char *path, unsigned mode)
+int dir::create(const char *path, unsigned perms)
 {
-    if(::mkdir(path, mode))
+    if(perms & 06)
+        perms |= 01;
+    if(perms & 060)
+        perms |= 010;
+    if(perms & 0600)
+        perms |= 0100;
+
+    if(::mkdir(path, perms))
         return remapError();
     return 0;
 }
 
-int fsys::removeDir(const char *path)
+void dir::open(const char *path)
 {
-    if(::rmdir(path))
-        return remapError();
-    return 0;
+    close();
+
+    ptr = opendir(path);
+    if(!ptr)
+        error = remapError();
 }
 
 void fsys::open(const char *path, access_t access)
@@ -804,46 +921,48 @@ void fsys::open(const char *path, access_t access)
 
     switch(access)
     {
-    case ACCESS_STREAM:
+    case STREAM:
 #if defined(O_STREAMING)
         flags = O_RDONLY | O_STREAMING;
         break;
 #endif
-    case ACCESS_RDONLY:
+    case RDONLY:
         flags = O_RDONLY;
         break;
-    case ACCESS_WRONLY:
+    case WRONLY:
         flags = O_WRONLY;
         break;
-    case ACCESS_RANDOM:
-    case ACCESS_SHARED:
-    case ACCESS_REWRITE:
-        flags = O_RDWR;
+    case EXCLUSIVE:
+    case RANDOM:
+    case SHARED:
+    case REWRITE:
+    case DEVICE:
+        flags = O_RDWR | O_NONBLOCK;
         break;
-    case ACCESS_APPEND:
+    case APPEND:
         flags = O_RDWR | O_APPEND;
         break;
-    case ACCESS_DIRECTORY:
-        ptr = opendir(path);
-        if(!ptr)
-            error = remapError();
-        return;
     }
     fd = ::open(path, flags);
-    if(fd == INVALID_HANDLE_VALUE)
+    if(fd == INVALID_HANDLE_VALUE) {
         error = remapError();
-#ifdef HAVE_POSIX_FADVISE
-    else {
-        // Linux kernel bug prevents use of POSIX_FADV_NOREUSE in streaming...
-        if(access == ACCESS_STREAM)
-            posix_fadvise(fd, (off_t)0, (off_t)0, POSIX_FADV_SEQUENTIAL);
-        else if(access == ACCESS_RANDOM)
-            posix_fadvise(fd, (off_t)0, (off_t)0, POSIX_FADV_RANDOM);
+        return;
     }
+#ifdef HAVE_POSIX_FADVISE
+    // Linux kernel bug prevents use of POSIX_FADV_NOREUSE in streaming...
+    if(access == STREAM)
+        posix_fadvise(fd, (off_t)0, (off_t)0, POSIX_FADV_SEQUENTIAL);
+    else if(access == RANDOM)
+        posix_fadvise(fd, (off_t)0, (off_t)0, POSIX_FADV_RANDOM);
 #endif
+    if(access == DEVICE) {
+        flags = fcntl(fd, F_GETFL);
+        flags &= ~O_NONBLOCK;
+        fcntl(fd, F_SETFL, flags);
+    }
 }
 
-int fsys::fileinfo(const char *path, struct stat *ino)
+int fsys::info(const char *path, struct stat *ino)
 {
     if(::stat(path, ino))
         return remapError();
@@ -870,7 +989,7 @@ int fsys::trunc(offset_t offset)
 }
 #endif
 
-int fsys::fileinfo(struct stat *ino)
+int fsys::info(struct stat *ino)
 {
     if(::fstat(fd, ino)) {
         error = remapError();
@@ -879,46 +998,72 @@ int fsys::fileinfo(struct stat *ino)
     return 0;
 }
 
-int fsys::changeDir(const char *path)
+int fsys::prefix(const char *path)
 {
     if(::chdir(path))
         return remapError();
     return 0;
 }
 
-int fsys::getPrefix(char *path, size_t len)
+int fsys::prefix(char *path, size_t len)
 {
     if(::getcwd(path, len))
         return remapError();
     return 0;
 }
 
-int fsys::change(const char *path, unsigned mode)
+int fsys::mode(const char *path, unsigned value)
 {
-    if(::chmod(path, mode))
+    if(::chmod(path, value))
         return remapError();
     return 0;
 }
 
-int fsys::access(const char *path, unsigned mode)
+bool fsys::is_exists(const char *path)
 {
-    if(::access(path, mode))
-        return remapError();
-    return 0;
+    if(::access(path, F_OK))
+        return false;
+
+    return true;
+}
+
+bool fsys::is_readable(const char *path)
+{
+    if(::access(path, R_OK))
+        return false;
+
+    return true;
+}
+
+bool fsys::is_writable(const char *path)
+{
+    if(::access(path, W_OK))
+        return false;
+
+    return true;
+}
+
+bool fsys::is_executable(const char *path)
+{
+    if(is_dir(path))
+        return false;
+
+    if(::access(path, X_OK))
+        return false;
+
+    return true;
 }
 
 fsys::fsys(const fsys& copy)
 {
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
     error = 0;
 
-    if(copy.fd != INVALID_HANDLE_VALUE)
+    if(copy.fd != INVALID_HANDLE_VALUE) {
         fd = ::dup(copy.fd);
+    }
     else
         fd = INVALID_HANDLE_VALUE;
-    error = 0;
-    ptr = NULL;
 }
 
 int fsys::inherit(fd_t& fd, bool enable)
@@ -988,28 +1133,60 @@ int fsys::seek(offset_t pos)
 
 #endif
 
+dso::dso()
+{
+    ptr = 0;
+    error = 0;
+}
+
+dso::dso(const char *path)
+{
+    ptr = 0;
+    error = 0;
+    map(path);
+}
+
+dso::~dso()
+{
+    release();
+}
+
+dir::dir() :
+fsys()
+{
+    ptr = NULL;
+}
+
+dir::~dir()
+{
+    close();
+}
+
 fsys::fsys()
 {
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
     error = 0;
 }
 
+dir::dir(const char *path) :
+fsys()
+{
+    ptr = NULL;
+    open(path);
+}
+
 fsys::fsys(const char *path, access_t access)
 {
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
     error = 0;
     open(path, access);
 }
 
-
-fsys::fsys(const char *path, access_t access, unsigned mode)
+fsys::fsys(const char *path, unsigned fmode, access_t access)
 {
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
     error = 0;
-    create(path, access, mode);
+    open(path, fmode, access);
 }
 
 fsys::~fsys()
@@ -1032,7 +1209,7 @@ int fsys::linkinfo(const char *path, char *buffer, size_t size)
     char reparse[MAXIMUM_REPARSE_DATA_BUFFER_SIZE];
     DWORD rsize;
 
-    if(!fsys::islink(path))
+    if(!fsys::is_link(path))
         return EINVAL;
 
     h = CreateFile(path, GENERIC_READ, 0, 0, OPEN_EXISTING,
@@ -1128,7 +1305,7 @@ int fsys::unlink(const char *path)
 {
 #ifdef  _MSWINDOWS_
     HANDLE h = INVALID_HANDLE_VALUE;
-    if(islink(path))
+    if(is_link(path))
         h = CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING,
             FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, 0);
     if(!h || h != INVALID_HANDLE_VALUE) {
@@ -1146,11 +1323,39 @@ int fsys::unlink(const char *path)
         return 0;
     }
 #endif
-    return fsys::remove(path);
+    if(::remove(path))
+        return remapError();
+    return 0;
 }
 
-int fsys::remove(const char *path)
+int fsys::erase(const char *path)
 {
+    if(is_device(path))
+        return ENOSYS;
+
+    if(::remove(path))
+        return remapError();
+    return 0;
+}
+
+int dir::remove(const char *path)
+{
+    if(is_device(path))
+        return ENOSYS;
+
+#ifdef  _MSWINDOWS_
+    if(RemoveDirectory(path))
+        return 0;
+    int error = remapError();
+    if(error == ENOTEMPTY)
+        return ENOTEMPTY;
+#else
+    if(!::rmdir(path))
+        return 0;
+    if(errno != ENOTDIR)
+        return errno;
+#endif
+
     if(::remove(path))
         return remapError();
     return 0;
@@ -1170,11 +1375,11 @@ int fsys::copy(const char *oldpath, const char *newpath, size_t size)
 
     remove(newpath);
 
-    src.open(oldpath, fsys::ACCESS_STREAM);
+    src.open(oldpath, fsys::STREAM);
     if(!is(src))
         goto end;
 
-    dest.create(newpath, fsys::ACCESS_STREAM, 0664);
+    dest.open(newpath, GROUP_PUBLIC, fsys::STREAM);
     if(!is(dest))
         goto end;
 
@@ -1217,12 +1422,12 @@ int fsys::rename(const char *oldpath, const char *newpath)
 
 int fsys::load(const char *path)
 {
-    fsys module;
+    dso module;
 
-    load(module, path);
+    module.map(path);
 #ifdef  _MSWINDOWS_
-    if(module.mem) {
-        module.mem = 0;
+    if(module.ptr) {
+        module.ptr = 0;
         return 0;
     }
     return remapError();
@@ -1235,7 +1440,7 @@ int fsys::load(const char *path)
 #endif
 }
 
-bool fsys::isfile(const char *path)
+bool fsys::is_file(const char *path)
 {
 #ifdef _MSWINDOWS_
     DWORD attr = GetFileAttributes(path);
@@ -1260,7 +1465,7 @@ bool fsys::isfile(const char *path)
 #endif
 }
 
-bool fsys::islink(const char *path)
+bool fsys::is_link(const char *path)
 {
 #if defined(_MSWINDOWS_)
     DWORD attr = GetFileAttributes(path);
@@ -1282,7 +1487,7 @@ bool fsys::islink(const char *path)
 #endif
 }
 
-bool fsys::isdir(const char *path)
+bool fsys::is_dir(const char *path)
 {
 #ifdef _MSWINDOWS_
     DWORD attr = GetFileAttributes(path);
@@ -1309,24 +1514,27 @@ bool fsys::isdir(const char *path)
 
 #ifdef  _MSWINDOWS_
 
-void fsys::load(fsys& module, const char *path)
+void dso::map(const char *path)
 {
-    module.error = 0;
-    module.mem = LoadLibrary(path);
-    if(!module.mem)
-        module.error = ENOEXEC;
+    error = 0;
+    ptr = LoadLibrary(path);
+    if(!ptr)
+        error = ENOEXEC;
 }
 
-void fsys::unload(fsys& module)
+void dso::release(void)
 {
-    if(module.mem)
-        FreeLibrary(module.mem);
-    module.mem = 0;
+    if(ptr)
+        FreeLibrary(ptr);
+    ptr = 0;
 }
 
-void *fsys::find(fsys& module, const char *sym)
+dso::addr_t dso::find(const char *sym) const
 {
-    return (void *)GetProcAddress(module.mem, sym);
+    if(ptr == 0)
+        return (dso::addr_t)NULL;
+
+    return (addr_t)GetProcAddress(ptr, sym);
 }
 
 #elif defined(HAVE_DLFCN_H)
@@ -1336,51 +1544,51 @@ void *fsys::find(fsys& module, const char *sym)
 #define RTLD_GLOBAL 0
 #endif
 
-void fsys::load(fsys& module, const char *path)
+void dso::map(const char *path)
 {
-    module.error = 0;
-    module.ptr = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
-    if(module.ptr == NULL)
-        module.error = ENOEXEC;
+    error = 0;
+    ptr = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
+    if(ptr == NULL)
+        error = ENOEXEC;
 }
 
-void fsys::unload(fsys& module)
+void dso::release(void)
 {
-    if(module.ptr)
-        dlclose(module.ptr);
-    module.ptr = NULL;
+    if(ptr)
+        dlclose(ptr);
+    ptr = NULL;
 }
 
-void *fsys::find(fsys& module, const char *sym)
+dso::addr_t dso::find(const char *sym) const
 {
-    if(!module.ptr)
-        return NULL;
+    if(!ptr)
+        return (dso::addr_t)NULL;
 
-    return dlsym(module.ptr, (char *)sym);
+    return (dso::addr_t)dlsym(ptr, (char *)sym);
 }
 
 #elif HAVE_MACH_O_DYLD_H
 #include <mach-o/dyld.h>
 
-void fsys::load(fsys& module, const char *path)
+void dso::map(const char *path)
 {
     NSObjectFileImage oImage;
     NSSymbol sym = NULL;
     NSModule mod;
     void (*init)(void);
 
-    module.ptr = NULL;
-    module.error = 0;
+    ptr = NULL;
+    error = 0;
 
     if(NSCreateObjectFileImageFromFile(path, &oImage) != NSObjectFileImageSuccess) {
-        module.error = ENOEXEC;
+        error = ENOEXEC;
         return;
     }
 
     mod = NSLinkModule(oImage, path, NSLINKMODULE_OPTION_BINDNOW | NSLINKMODULE_OPTION_RETURN_ON_ERROR);
     NSDestroyObjectFileImage(oImage);
     if(mod == NULL) {
-        module.error = ENOEXEC;
+        error = ENOEXEC;
         return;
     }
 
@@ -1389,18 +1597,18 @@ void fsys::load(fsys& module, const char *path)
         init = (void (*)(void))NSAddressOfSymbol(sym);
         init();
     }
-    module.ptr = (void *)mod;
+    ptr = (void *)mod;
 }
 
-void fsys::unload(fsys& module)
+void dso::release(void)
 {
-    if(!module.ptr)
+    if(!ptr)
         return;
 
-    NSModule mod = (NSModule)module.ptr;
+    NSModule mod = (NSModule)ptr;
     NSSymbol sym;
     void (*fini)(void);
-    module.ptr = NULL;
+    ptr = NULL;
 
     sym = NSLookupSymbolInModule(mod, "__fini");
     if(sym != NULL) {
@@ -1410,70 +1618,116 @@ void fsys::unload(fsys& module)
     NSUnlinkModule(mod, NSUNLINKMODULE_OPTION_NONE);
 }
 
-void *fsys::find(fsys& module, const char *sym)
+dso::addr_t dso::find(const char *sym) const
 {
-    if(!module.ptr)
+    if(!ptr)
         return NULL;
 
-    NSModule mod = (NSModule)module.ptr;
+    NSModule mod = (NSModule)ptr;
     NSSymbol sym;
 
     sym = NSLookupSymbolInModule(mod, sym);
     if(sym != NULL) {
-        return NSAddressOfSymbol(sym);
+        return (dso::addr_t)NSAddressOfSymbol(sym);
 
-    return NULL;
+    return (dso::addr_t)NULL;
 }
 
 #elif HAVE_SHL_LOAD
 #include <dl.h>
 
-void fsys::load(fsys& module, const char *path)
+void dso::map(const char *path)
 {
-    module.error = 0;
-    module.ptr = (void *))shl_load(path, BIND_IMMEDIATE, 0l);
-    if(!module.ptr)
-        module.error = ENOEXEC;
+    error = 0;
+    ptr = (void *)shl_load(path, BIND_IMMEDIATE, 0l);
+    if(!ptr)
+        error = ENOEXEC;
 }
 
-void *fsys::find(fsys& module, const char *sym)
+dso::addr_t dso::find(const char *sym) const
 {
-    shl_t image = (shl_t)module.ptr;
+    if(!ptr)
+        return (dso::addr_t)NULL;
+
+    shl_t image = (shl_t)ptr;
 
     if(shl_findsym(&image, sym, 0, &value) == 0)
-        return (void *)value;
+        return (dso::addr_t)value;
 
-    return NULL;
+    return (dso::addr_t)NULL;
 }
 
-void fsys::unload(fsys& module)
+void dso::release(void)
 {
-    shl_t image = (shl_t)module.ptr;
-    if(addr)
+    shl_t image = (shl_t)ptr;
+    if(ptr)
         shl_unload(image);
-    module.ptr = NULL;
+    ptr = NULL;
 }
 
 #else
 
-void fsys::load(fsys& module, const char *path)
+void fsys::map(const char *path)
 {
-    module.error = ENOEXEC;
-    module.ptr = NULL;
+    error = ENOEXEC;
+    ptr = NULL;
 }
 
-void fsys::unload(mem_t addr)
+void dso::release(void)
 {
 }
 
-void *fsys::find(mem_t addr, const char *sym)
+dso::addr_t dso::find(const char *sym) const
 {
-    return NULL;
+    return (dso::addr_t)NULL;
 }
 
 #endif
 
-bool fsys::ishidden(const char *path)
+bool fsys::is_device(const char *path)
+{
+    if(!path)
+        return false;
+
+#ifndef _MSWINDOWS_
+    if(is_dir(path))
+        return false;
+
+    if(!strncmp(path, "/dev/", 5))
+        return true;
+
+    return false;
+#else
+    if(path[1] == ':' && !path[2] && isalpha(*path))
+        return true;
+
+    if(!strncmp(path, "com", 3) || !strncmp(path, "lpt", 3)) {
+        path += 3;
+        while(isdigit(*path))
+            ++path;
+        if(!path || *path == ':')
+            return true;
+        return false;
+    }
+
+    if(!strcmp(path, "aux") || !strcmp(path, "prn")) {
+        if(!path[3] || path[3] == ':')
+            return true;
+        return false;
+    }
+
+    if(!strncmp(path, "\\\\.\\", 4))
+        return true;
+
+    if(!strnicmp(path, "\\\\?\\Device\\", 12))
+        return true;
+
+    return false;
+
+#endif
+}
+
+bool fsys::is_hidden(const char *path)
 {
 #ifdef  _MSWINDOWS_
     DWORD attr = GetFileAttributes(path);
@@ -1495,164 +1749,9 @@ bool fsys::ishidden(const char *path)
 #endif
 }
 
-charfile::charfile(const char *file, const char *mode)
-{
-    fp = NULL;
-    open(file, mode);
-}
-
-charfile::charfile()
-{
-    fp = NULL;
-    opened = false;
-}
-
-charfile::~charfile()
-{
-    close();
-}
-
-bool charfile::istty(void)
-{
-#ifdef  _MSWINDOWS_
-    if(_isatty(_fileno(fp)))
-        return true;
-#else
-    if(isatty(fileno(fp)))
-        return true;
-#endif
-    return false;
-}
-
-void charfile::open(const char *path, const char *mode)
-{
-    if(fp)
-        fclose(fp);
-
-    fp = fopen(path, mode);
-    opened = true;
-}
-
-void charfile::close(void)
-{
-    if(fp && opened)
-        fclose(fp);
-    fp = NULL;
-}
-
-size_t charfile::printf(const char *format, ...)
-{
-    if(!fp)
-        return 0;
-
-    va_list args;
-    va_start(args, format);
-    size_t result = vfprintf(fp, format, args);
-    va_end(args);
-    if(result == (size_t)EOF)
-        result = 0;
-    return result;
-}
-
-size_t charfile::put(const char *data)
-{
-    if(!fp)
-        return 0;
-
-    int result = fputs(data, fp);
-    if(result < 0)
-        result = 0;
-
-    return (size_t)result;
-}
-
-size_t charfile::readline(char *address, size_t size)
-{
-    address[0] = 0;
-
-    if(!fp)
-        return false;
-
-    if(!fgets(address, size, fp) || feof(fp))
-        return 0;
-
-    size_t result = size = strlen(address);
-
-    if(address[size - 1] == '\n') {
-        --size;
-        if(size && address[size - 1] == '\r')
-            --size;
-    }
-    address[size] = 0;
-    return result;
-}
-
-size_t charfile::readline(string& s)
-{
-    if(!s.c_mem())
-        return true;
-
-    if(!fgets(s.c_mem(), s.size(), fp) || feof(fp)) {
-        s.clear();
-        return false;
-    }
-
-    String::fix(s);
-    size_t result = s.len();
-
-    if(s[-1] == '\n')
-        --s;
-
-    if(s[-1] == '\r')
-        --s;
-
-    return result;
-}
-
-bool charfile::eof(void)
-{
-    if(!fp)
-        return false;
-
-    return feof(fp) != 0;
-}
-
-int charfile::err(void)
-{
-    if(!fp)
-        return EBADF;
-
-    return ferror(fp);
-}
-
-int charfile::_getch(void)
-{
-    if(!fp)
-        return EOF;
-
-    return fgetc(fp);
-}
-
-int charfile::_putch(int code)
-{
-    if(!fp)
-        return EOF;
-
-    return fputc(code, fp);
-}
-
-String str(charfile& so, strsize_t size)
-{
-    String s(size);
-    so.readline(s.c_mem(), s.size());
-    String::fix(s);
-    return s;
-}
-
 fsys::fsys(fd_t handle)
 {
     fd = handle;
-    ptr = NULL;
     error = 0;
 }
 
@@ -1660,7 +1759,6 @@ void fsys::set(fd_t handle)
 {
     close();
     fd = handle;
-    ptr = NULL;
     error = 0;
 }
 
@@ -1669,20 +1767,21 @@ fd_t fsys::release(void)
     fd_t save = fd;
 
     fd = INVALID_HANDLE_VALUE;
-    ptr = NULL;
     error = 0;
     return save;
 }
 
-int fsys::exec(const char *path, char **argv)
-{
-    shell::pid_t pid = shell::spawn(path, argv);
-    return shell::wait(pid);
-}
-
 int fsys::exec(const char *path, char **argv, char **envp)
 {
     shell::pid_t pid = shell::spawn(path, argv, envp);
     return shell::wait(pid);
 }
 
+string_t fsys::prefix(void)
+{
+    char buf[256];
+
+    prefix(buf, sizeof(buf));
+    string_t s = buf;
+    return s;
+}
