@@ -20,20 +20,34 @@
 #include <ucommon/export.h>
 #include <ucommon/protocols.h>
 #include <ucommon/string.h>
+#include <ucommon/memory.h>
 #include <stdlib.h>
+#include <ctype.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
+#undef  getc
+#undef  putc
+#undef  puts
+#undef  gets
+
 using namespace UCOMMON_NAMESPACE;
 
+void MemoryProtocol::fault(void) const
+{
+}
+
 char *MemoryProtocol::dup(const char *str)
 {
     if(!str)
         return NULL;
     size_t len = strlen(str) + 1;
     char *mem = static_cast<char *>(alloc(len));
-    String::set(mem, len, str);
+    if(mem)
+        String::set(mem, len, str);
+    else
+        fault();
     return mem;
 }
 
@@ -43,7 +57,10 @@ void *MemoryProtocol::dup(void *obj, size_t size)
     assert(size > 0);
 
     char *mem = static_cast<char *>(alloc(size));
-    memcpy(mem, obj, size);
+    if(mem)
+        memcpy(mem, obj, size);
+    else
+        fault();
     return mem;
 }
 
@@ -53,6 +70,8 @@ void *MemoryProtocol::zalloc(size_t size)
 
     if(mem)
         memset(mem, 0, size);
+    else
+        fault();
 
     return mem;
 }
@@ -75,14 +94,14 @@ void *MemoryRedirect::_alloc(size_t size)
     return target->_alloc(size);
 }
 
-BufferProtocol::BufferProtocol()
+BufferProtocol::BufferProtocol() : CharacterProtocol()
 {
     end = true;
     eol = "\r\n";
     input = output = buffer = NULL;
 }
 
-BufferProtocol::BufferProtocol(size_t size, type_t mode)
+BufferProtocol::BufferProtocol(size_t size, mode_t mode)
 {
     end = true;
     eol = "\r\n";
@@ -95,6 +114,10 @@ BufferProtocol::~BufferProtocol()
     release();
 }
 
+void BufferProtocol::fault(void) const
+{
+}
+
 void BufferProtocol::release(void)
 {
     if(buffer) {
@@ -105,7 +128,7 @@ void BufferProtocol::release(void)
     }
 }
 
-void BufferProtocol::allocate(size_t size, type_t mode)
+void BufferProtocol::allocate(size_t size, mode_t mode)
 {
     release();
     _clear();
@@ -113,17 +136,24 @@ void BufferProtocol::allocate(size_t size, type_t mode)
     if(!size)
         return;
 
+    back = 0;
     switch(mode) {
-    case BUF_RDWR:
+    case RDWR:
         input = buffer = (char *)malloc(size * 2);
         if(buffer)
             output = buffer + size;
+        else
+            fault();
         break;
-    case BUF_RD:
+    case RDONLY:
         input = buffer = (char *)malloc(size);
+        if(!buffer)
+            fault();
         break;
-    case BUF_WR:
+    case WRONLY:
         output = buffer = (char *)malloc(size);
+        if(!buffer)
+            fault();
         break;
     }
 
@@ -139,13 +169,38 @@ bool BufferProtocol::_blocking(void)
     return false;
 }
 
-size_t BufferProtocol::get(char *address, size_t size)
+size_t BufferProtocol::put(const void *address, size_t size)
 {
     size_t count = 0;
 
-    if(!input || !address)
+    if(!output || !address || !size)
         return 0;
 
+    const char *cp = (const char *)address;
+
+    while(count < size) {
+        if(outsize == bufsize) {
+            outsize = 0;
+            if(_push(output, bufsize) < bufsize) {
+                output = NULL;
+                end = true;     // marks a disconnection...
+                return count;
+            }
+        }
+        output[outsize++] = cp[count++];
+    }
+    return count;
+}
+
+size_t BufferProtocol::get(void *address, size_t size)
+{
+    size_t count = 0;
+
+    if(!input || !address || !size)
+        return 0;
+
+    char *cp = (char *)address;
+
     while(count < size) {
         if(bufpos == insize) {
             if(end)
@@ -161,7 +216,7 @@ size_t BufferProtocol::get(char *address, size_t size)
             if(!insize)
                 return count;
         }
-        address[count++] = input[bufpos++];
+        cp[count++] = input[bufpos++];
     }
     return count;
 }
@@ -171,6 +226,11 @@ int BufferProtocol::_getch(void)
     if(!input)
         return EOF;
 
+    if(back) {
+        back = 0;
+        return back;
+    }
+
     if(bufpos == insize) {
         if(end)
             return EOF;
@@ -189,27 +249,21 @@ int BufferProtocol::_getch(void)
     return input[bufpos++];
 }
 
-size_t BufferProtocol::put(const char *address, size_t size)
+size_t CharacterProtocol::putchars(const char *address, size_t size)
 {
     size_t count = 0;
 
-    if(!output || !address)
+    if(!address)
         return 0;
 
     if(!size)
         size = strlen(address);
 
     while(count < size) {
-        if(outsize == bufsize) {
-            outsize = 0;
-            if(_push(output, bufsize) < bufsize) {
-                output = NULL;
-                end = true;     // marks a disconnection...
-                return count;
-            }
-        }
-
-        output[outsize++] = address[count++];
+        if(_putch(*address) == EOF)
+            break;
+        ++count;
+        ++address;
     }
 
     return count;
@@ -220,6 +274,12 @@ int BufferProtocol::_putch(int ch)
     if(!output)
         return EOF;
 
+    if(ch == 0) {
+        puts(eol);
+        flush();
+        return 0;
+    }
+
     if(outsize == bufsize) {
         outsize = 0;
         if(_push(output, bufsize) < bufsize) {
@@ -233,17 +293,17 @@ int BufferProtocol::_putch(int ch)
     return ch;
 }
 
-size_t BufferProtocol::printf(const char *format, ...)
+size_t BufferProtocol::printf(const char *pformat, ...)
 {
     va_list args;
     int result;
     size_t count;
 
-    if(!flush() || !output || !format)
+    if(!flush() || !output || !pformat)
         return 0;
 
-    va_start(args, format);
-    result = vsnprintf(output, bufsize, format, args);
+    va_start(args, pformat);
+    result = vsnprintf(output, bufsize, pformat, args);
     va_end(args);
     if(result < 1)
         return 0;
@@ -322,27 +382,28 @@ char *BufferProtocol::request(size_t size)
     return output + offset;
 }
 
-size_t BufferProtocol::putline(const char *string)
+size_t CharacterProtocol::putline(const char *string)
 {
     size_t count = 0;
 
-    if(string)
-        count += put(string);
+    while(string && *string && (EOF != _putch(*string)))
+        ++count;
 
-    if(eol)
-        count += put(eol);
+    string = eol;
+    while(string && *string && (EOF != _putch(*string)))
+        ++count;
 
     return count;
 }
 
-size_t BufferProtocol::getline(string& s)
+size_t CharacterProtocol::getline(String& s)
 {
     size_t result = getline(s.c_mem(), s.size() + 1);
     String::fix(s);
     return result;
 }
 
-size_t BufferProtocol::getline(char *string, size_t size)
+size_t CharacterProtocol::getline(char *string, size_t size)
 {
     size_t count = 0;
     unsigned eolp = 0;
@@ -350,16 +411,13 @@ size_t BufferProtocol::getline(char *string, size_t size)
     bool eof = false;
 
     if(!eols)
-        eols="\0";
+        eols = "\0";
 
     if(string)
         string[0] = 0;
 
-    if(!input || !string)
-        return 0;
-
     while(count < size - 1) {
-        int ch = BufferProtocol::_getch();
+        int ch = _getch();
         if(ch == EOF) {
             eolp = 0;
             eof = true;
@@ -389,6 +447,75 @@ size_t BufferProtocol::getline(char *string, size_t size)
     return count;
 }
 
+size_t CharacterProtocol::print(const PrintProtocol& f)
+{
+    const char *cp = f._print();
+
+    if(cp == NULL)
+        return putchar(0);
+
+    return putchars(cp);
+}
+
+size_t CharacterProtocol::input(InputProtocol& f)
+{
+    int c;
+    size_t count = 0;
+
+    for(;;) {
+        if(back) {
+            c = back;
+            back = 0;
+        }
+        else
+            c = _getch();
+
+        c = f._input(c);
+        if(c) {
+            if(c != EOF)
+                back = c;
+            else
+                ++count;
+            break;
+        }
+        ++count;
+    }
+    return count;
+}
+
+size_t CharacterProtocol::load(StringPager *list)
+{
+    if(!list)
+        return 0;
+
+    size_t used = 0;
+    size_t size = list->size() - 64;
+
+    char *tmp = (char *)malloc(size);
+    while(getline(tmp, size)) {
+        if(!list->filter(tmp, size))
+            break;
+
+        ++used;
+    }
+    free(tmp);
+    return used;
+}
+
+size_t CharacterProtocol::save(const StringPager *list)
+{
+    size_t used = 0;
+    if(!list)
+        return 0;
+
+    StringPager::iterator sp = list->begin();
+    while(is(sp) && putline(sp->get())) {
+        ++used;
+        sp.next();
+    }
+    return used;
+}
+
 void BufferProtocol::reset(void)
 {
     insize = bufpos = 0;
@@ -428,6 +555,187 @@ LockingProtocol::~LockingProtocol()
 {
 }
 
+CharacterProtocol::CharacterProtocol()
+{
+    back = 0;
+    eol = "\n";
+}
+
 CharacterProtocol::~CharacterProtocol()
 {
 }
+
+ObjectProtocol::~ObjectProtocol()
+{
+}
+
+ObjectProtocol *ObjectProtocol::copy(void)
+{
+    retain();
+    return this;
+}
+
+CharacterProtocol& _character_operators::print(CharacterProtocol& p, const char& c)
+{
+    p.putchar((int)c);
+    return p;
+}
+
+CharacterProtocol& _character_operators::input(CharacterProtocol& p, char& c)
+{
+    int code = p.getchar();
+    if(code == EOF)
+        code = 0;
+    c = code;
+    return p;
+}
+
+CharacterProtocol& _character_operators::print(CharacterProtocol& p, const char *str)
+{
+    if(!str)
+        p.putline("");
+    else
+        p.putchars(str);
+    return p;
+}
+
+CharacterProtocol& _character_operators::input(CharacterProtocol& p, String& str)
+{
+    if(str.c_mem()) {
+        p.getline(str.c_mem(), str.size());
+        String::fix(str);
+    }
+    return p;
+}
+
+CharacterProtocol& _character_operators::print(CharacterProtocol& p, const long& v)
+{
+    char buf[40];
+    snprintf(buf, sizeof(buf), "%ld", v);
+    p.putchars(buf);
+    return p;
+}
+
+CharacterProtocol& _character_operators::print(CharacterProtocol& p, const double& v)
+{
+    char buf[40];
+    snprintf(buf, sizeof(buf), "%f", v);
+    p.putchars(buf);
+    return p;
+}
+
+
+class __LOCAL _input_long : public InputProtocol
+{
+public:
+    long* ref;
+    size_t pos;
+    char buf[32];
+
+    _input_long(long& v);
+
+    int _input(int code);
+};
+
+class __LOCAL _input_double : public InputProtocol
+{
+public:
+    double* ref;
+    bool dot;
+    bool e;
+    size_t pos;
+    char buf[60];
+
+    _input_double(double& v);
+
+    int _input(int code);
+};
+
+_input_long::_input_long(long& v)
+{
+    ref = &v;
+    v = 0l;
+    pos = 0;
+}
+
+_input_double::_input_double(double& v)
+{
+    dot = e = false;
+    v = 0.0;
+    pos = 0;
+    ref = &v;
+}
+
+int _input_long::_input(int code)
+{
+    if(code == '-' && !pos)
+        goto valid;
+
+    if(isdigit(code) && pos < sizeof(buf) - 1)
+        goto valid;
+
+    buf[pos] = 0;
+    if(pos)
+        sscanf(buf, "%ld", ref);
+
+    return code;
+
+valid:
+    buf[pos++] = code;
+    return 0;
+}
+
+int _input_double::_input(int code)
+{
+    if(code == '-' && !pos)
+        goto valid;
+
+    if(code == '-' && buf[pos] == 'e')
+        goto valid;
+
+    if(tolower(code) == 'e' && !e) {
+        e = true;
+        code = 'e';
+        goto valid;
+    }
+
+    if(code == '.' && !dot) {
+        dot = true;
+        goto valid;
+    }
+
+    if(isdigit(code) && pos < sizeof(buf) - 1)
+        goto valid;
+
+    buf[pos] = 0;
+    if(pos)
+        sscanf(buf, "%lf", ref);
+
+    return code;
+
+valid:
+    buf[pos++] = code;
+    return 0;
+}
+
+CharacterProtocol& _character_operators::input(CharacterProtocol& p, long& v)
+{
+    _input_long lv(v);
+    p.input(lv);
+    return p;
+}
+
+CharacterProtocol& _character_operators::input(CharacterProtocol& p, double& v)
+{
+    _input_double lv(v);
+    p.input(lv);
+    return p;
+}
+
+PrintProtocol::~PrintProtocol()
+{
+}
+
+InputProtocol::~InputProtocol()
+{
+}
