@@ -30,7 +30,7 @@ static bool temp = false;
 static void cleanup(void)
 {
     if(temp)
-        fsys::remove("zerofill.tmp");
+        fsys::erase("zerofill.tmp");
     if(live)
         shell::printf("\n");
     live = false;
@@ -45,7 +45,7 @@ static void zerofill(void)
 
     temp = true;
 
-    fs.create("zerofill.tmp", fsys::ACCESS_STREAM, 0666);
+    fs.open("zerofill.tmp", 0666, fsys::STREAM);
     if(!is(fs))
         shell::errexit(1, "*** zerofill: %s\n",
             _TEXT("cannot create temporary file"));
@@ -86,15 +86,15 @@ static void zerofill(const char *devname)
     unsigned long count = 0;
     fsys_t fs;
 
-    if(fsys::fileinfo(devname, &ino))
+    if(fsys::info(devname, &ino))
         shell::errexit(5, "*** zerofill: %s: %s\n",
             devname, _TEXT("cannot access"));
 
-    if(!fsys::isdisk(&ino))
+    if(!fsys::is_disk(&ino))
         shell::errexit(6, "*** zerofill: %s: %s\n",
             devname, _TEXT("not block device"));
 
-    fs.open(devname, fsys::ACCESS_WRONLY);
+    fs.open(devname, fsys::WRONLY);
     if(fs.err())
         shell::errexit(5, "*** zerofill: %s: %s\n",
             devname, _TEXT("cannot modify"));
